import json
from abc import ABC, abstractmethod
from typing import Any, Dict

from kilroy_face_server_py_sdk import (
    Categorizable,
    ImageData,
    ImageOnlyPost,
    ImageWithOptionalTextPost,
    JSONSchema,
    TextAndImagePost,
    TextData,
    TextOnlyPost,
    TextOrImagePost,
    TextWithOptionalImagePost,
    classproperty,
    normalize,
)

from kilroy_face_debug.post import PostData, PostTextData, PostImageData


class Processor(Categorizable, ABC):
    # noinspection PyMethodParameters
    @classproperty
    def category(cls) -> str:
        name: str = cls.__name__
        return normalize(name.removesuffix("Processor"))

    @abstractmethod
    async def to_external(self, data: PostData) -> Dict[str, Any]:
        pass

    @abstractmethod
    async def to_internal(self, data: Dict[str, Any]) -> PostData:
        pass

    # noinspection PyMethodParameters
    @classproperty
    @abstractmethod
    def post_schema(cls) -> JSONSchema:
        pass


# Text only


class TextOnlyProcessor(Processor):
    # noinspection PyMethodParameters
    @classproperty
    def post_schema(cls) -> JSONSchema:
        return JSONSchema(**TextOnlyPost.schema())

    async def to_external(self, data: PostData) -> Dict[str, Any]:
        post = TextOnlyPost(text=TextData(content=data.text.content))
        return json.loads(post.json())

    async def to_internal(self, data: Dict[str, Any]) -> PostData:
        post = TextOnlyPost.parse_obj(data)
        return PostData(text=PostTextData(content=post.text.content))


# Image only


class ImageOnlyProcessor(Processor):
    # noinspection PyMethodParameters
    @classproperty
    def post_schema(cls) -> JSONSchema:
        return JSONSchema(**ImageOnlyPost.schema())

    async def to_external(self, data: PostData) -> Dict[str, Any]:
        post = ImageOnlyPost(
            image=ImageData(raw=data.image.raw, filename=data.image.filename)
        )
        return json.loads(post.json())

    async def to_internal(self, data: Dict[str, Any]) -> PostData:
        post = ImageOnlyPost.parse_obj(data)
        return PostData(
            image=PostImageData(
                raw=post.image.raw, filename=post.image.filename
            )
        )


# Text and image


class TextAndImageProcessor(Processor):
    # noinspection PyMethodParameters
    @classproperty
    def post_schema(cls) -> JSONSchema:
        return JSONSchema(**TextAndImagePost.schema())

    async def to_external(self, data: PostData) -> Dict[str, Any]:
        post = TextAndImagePost(
            text=TextData(content=data.text.content),
            image=ImageData(raw=data.image.raw, filename=data.image.filename),
        )
        return json.loads(post.json())

    async def to_internal(self, data: Dict[str, Any]) -> PostData:
        post = TextAndImagePost.parse_obj(data)
        return PostData(
            text=PostTextData(content=post.text.content),
            image=PostImageData(
                raw=post.image.raw, filename=post.image.filename
            ),
        )


# Text or image


class TextOrImageProcessor(Processor):
    # noinspection PyMethodParameters
    @classproperty
    def post_schema(cls) -> JSONSchema:
        return JSONSchema(**TextOrImagePost.schema())

    async def to_external(self, data: PostData) -> Dict[str, Any]:
        post = TextOrImagePost(
            text=TextData(content=data.text.content) if data.text else None,
            image=ImageData(raw=data.image.raw, filename=data.image.filename)
            if data.image
            else None,
        )
        return json.loads(post.json())

    async def to_internal(self, data: Dict[str, Any]) -> PostData:
        post = TextOrImagePost.parse_obj(data)
        text = (
            PostTextData(content=post.text.content)
            if post.text is not None
            else None
        )
        image = (
            PostImageData(raw=post.image.raw, filename=post.image.filename)
            if post.image is not None
            else None
        )
        return PostData(text=text, image=image)


# Text with optional image


class TextWithOptionalImageProcessor(Processor):
    # noinspection PyMethodParameters
    @classproperty
    def post_schema(cls) -> JSONSchema:
        return JSONSchema(**TextWithOptionalImagePost.schema())

    async def to_external(self, data: PostData) -> Dict[str, Any]:
        post = TextWithOptionalImagePost(
            text=TextData(content=data.text.content),
            image=ImageData(raw=data.image.raw, filename=data.image.filename)
            if data.image
            else None,
        )
        return json.loads(post.json())

    async def to_internal(self, data: Dict[str, Any]) -> PostData:
        post = TextWithOptionalImagePost.parse_obj(data)
        text = PostTextData(content=post.text.content)
        image = (
            PostImageData(raw=post.image.raw, filename=post.image.filename)
            if post.image is not None
            else None
        )
        return PostData(text=text, image=image)


# Image with optional text


class ImageWithOptionalTextProcessor(Processor):
    # noinspection PyMethodParameters
    @classproperty
    def post_schema(cls) -> JSONSchema:
        return JSONSchema(**ImageWithOptionalTextPost.schema())

    async def to_external(self, data: PostData) -> Dict[str, Any]:
        post = ImageWithOptionalTextPost(
            text=TextData(content=data.text.content) if data.text else None,
            image=ImageData(raw=data.image.raw, filename=data.image.filename),
        )
        return json.loads(post.json())

    async def to_internal(self, data: Dict[str, Any]) -> PostData:
        post = ImageWithOptionalTextPost.parse_obj(data)
        text = (
            PostTextData(content=post.text.content)
            if post.text is not None
            else None
        )
        image = PostImageData(raw=post.image.raw, filename=post.image.filename)
        return PostData(text=text, image=image)



from threading import Lock
from typing import Optional

from detoxify import Detoxify


def fetch_models() -> None:
    ToxicityModelLoader.get()
    ToxicityModelLoader.release()


class ToxicityModelLoader:
    model: Optional[Detoxify] = None
    reference_count: int = 0
    lock: Lock = Lock()

    @classmethod
    def get(cls) -> Detoxify:
        with cls.lock:
            if cls.model is None:
                cls.model = Detoxify("multilingual")
            cls.reference_count += 1
            return cls.model

    @classmethod
    def release(cls) -> None:
        with cls.lock:
            cls.reference_count -= 1
            if cls.reference_count == 0:
                cls.model = None



from abc import ABC, abstractmethod
from dataclasses import dataclass
from pathlib import Path
from random import random

from detoxify import Detoxify
from kilroy_face_server_py_sdk import Categorizable, classproperty, normalize
from kilroy_server_py_utils import Configurable, background

from kilroy_face_debug.models import ToxicityModelLoader
from kilroy_face_debug.post import Post


class Scorer(Categorizable, ABC):
    # noinspection PyMethodParameters
    @classproperty
    def category(cls) -> str:
        name: str = cls.__name__
        return normalize(name.removesuffix("Scorer"))

    @abstractmethod
    async def score(self, post: Post) -> float:
        pass


# Random


class RandomScorer(Scorer):
    async def score(self, post: Post) -> float:
        return random()


# Toxicity


@dataclass
class ToxicityScorerState:
    detoxify: Detoxify


class ToxicityScorer(Scorer, Configurable[ToxicityScorerState]):
    async def _build_default_state(self) -> ToxicityScorerState:
        return ToxicityScorerState(
            detoxify=await background(ToxicityModelLoader.get),
        )

    @classmethod
    async def _save_state(
        cls, state: ToxicityScorerState, directory: Path
    ) -> None:
        pass

    async def _load_saved_state(self, directory: Path) -> ToxicityScorerState:
        return await self._build_default_state()

    async def cleanup(self) -> None:
        await background(ToxicityModelLoader.release)

    async def score(self, post: Post) -> float:
        async with self.state.read_lock() as state:
            if not post.data.text:
                return 0.0
            return state.detoxify.predict(post.data.text.content)["toxicity"]



import json
from abc import ABC, abstractmethod
from dataclasses import dataclass
from pathlib import Path
from typing import Dict, Any

import numpy as np
from detoxify import Detoxify
from kilroy_face_py_shared import SerializableModel
from kilroy_face_server_py_sdk import Categorizable, classproperty, normalize
from kilroy_server_py_utils import Configurable, Parameter, background

from kilroy_face_debug.models import ToxicityModelLoader
from kilroy_face_debug.post import Post


class ScoreModifier(Categorizable, ABC):
    # noinspection PyMethodParameters
    @classproperty
    def category(cls) -> str:
        name: str = cls.__name__
        return normalize(name.removesuffix("ScoreModifier"))

    @abstractmethod
    async def modify(self, post: Post, score: float) -> float:
        pass


# Toxicity


class ToxicityScoreModifierParams(SerializableModel):
    threshold: float = 0.8
    alpha: float = 0.9


@dataclass
class ToxicityScoreModifierState:
    detoxify: Detoxify
    threshold: float
    alpha: float


class ToxicityScoreModifier(
    ScoreModifier, Configurable[ToxicityScoreModifierState]
):
    class ThresholdParameter(Parameter[ToxicityScoreModifierState, float]):
        # noinspection PyMethodParameters
        @classproperty
        def schema(cls) -> Dict[str, Any]:
            return {
                "type": "number",
                "minimum": 0,
                "maximum": 1,
                "title": cls.pretty_name,
                "default": 0.8,
            }

    class AlphaParameter(Parameter[ToxicityScoreModifierState, float]):
        # noinspection PyMethodParameters
        @classproperty
        def schema(cls) -> Dict[str, Any]:
            return {
                "type": "number",
                "minimum": 0,
                "maximum": 1,
                "title": cls.pretty_name,
                "default": 0.9,
            }

    async def _build_default_state(self) -> ToxicityScoreModifierState:
        params = ToxicityScoreModifierParams(**self._kwargs)
        return ToxicityScoreModifierState(
            detoxify=await background(ToxicityModelLoader.get),
            threshold=params.threshold,
            alpha=params.alpha,
        )

    @classmethod
    async def _save_state(
        cls, state: ToxicityScoreModifierState, directory: Path
    ) -> None:
        state_dict = {
            "threshold": state.threshold,
            "alpha": state.alpha,
        }
        with open(directory / "state.json", "w") as f:
            json.dump(state_dict, f)

    async def _load_saved_state(
        self, directory: Path
    ) -> ToxicityScoreModifierState:
        params = ToxicityScoreModifierParams(**self._kwargs)
        with open(directory / "state.json", "r") as f:
            state_dict = json.load(f)
        return ToxicityScoreModifierState(
            detoxify=await background(ToxicityModelLoader.get),
            threshold=state_dict.get("threshold", params.threshold),
            alpha=state_dict.get("alpha", params.alpha),
        )

    async def cleanup(self) -> None:
        await background(ToxicityModelLoader.release)

    async def modify(self, post: Post, score: float) -> float:
        async with self.state.read_lock() as state:
            content = post.data.text.content
            toxicity = state.detoxify.predict(content)["toxicity"]
            threshold = state.threshold
            alpha = state.alpha
        return self.modifier(toxicity, threshold, alpha) * score

    @staticmethod
    def modifier(x: float, threshold: float, alpha: float) -> float:
        x = np.clip(x, 0, 1).item()
        threshold = np.clip(threshold, 0, 1).item()
        alpha = np.clip(alpha, 0, 1).item()

        if x == 0 or x == 1:
            return 1 - x
        if threshold == 0:
            return 0
        if threshold == 1:
            return 1
        if alpha == 1 and x == threshold:
            return 1

        inner_exponent = (-np.log(2) / np.log(threshold)).item()
        outer_exponent = 1 / (1 - alpha)
        inner_value = x**inner_exponent
        denominator = 1 + (inner_value / (1 - inner_value)) ** outer_exponent
        return 1 / denominator



from abc import ABC, abstractmethod
from datetime import datetime
from typing import AsyncIterable, Optional, List, Dict, Any
from uuid import uuid4

from kilroy_face_server_py_sdk import (
    Categorizable,
    classproperty,
    normalize,
    SerializableState,
)
from kilroy_server_py_utils import Configurable, Parameter

from kilroy_face_debug.post import Post, PostTextData, PostData


class Scraper(Categorizable, ABC):
    # noinspection PyMethodParameters
    @classproperty
    def category(cls) -> str:
        name: str = cls.__name__
        return normalize(name.removesuffix("Scraper"))

    @abstractmethod
    def scrap(
        self,
        before: Optional[datetime] = None,
        after: Optional[datetime] = None,
    ) -> AsyncIterable[Post]:
        pass


# Dummy


class DummyScraperState(SerializableState):
    posts: List[str] = []


class DummyScraper(Scraper, Configurable[DummyScraperState]):
    class PostsParameter(Parameter[DummyScraperState, List[str]]):
        # noinspection PyMethodParameters
        @classproperty
        def schema(cls) -> Dict[str, Any]:
            return {
                "type": "array",
                "items": {"type": "string"},
                "title": cls.pretty_name,
                "default": [],
            }

    async def scrap(
        self,
        before: Optional[datetime] = None,
        after: Optional[datetime] = None,
    ) -> AsyncIterable[Post]:
        async with self.state.read_lock() as state:
            for post in state.posts:
                yield Post(
                    data=PostData(text=PostTextData(content=post)), id=uuid4()
                )



import json
import logging
from abc import ABC, abstractmethod
from dataclasses import dataclass
from datetime import datetime
from functools import partial
from pathlib import Path
from typing import Any, AsyncIterable, Dict, Optional, Set, Tuple, Type
from uuid import UUID

from aiostream import stream
from aiostream.aiter_utils import aiter, anext

from kilroy_face_debug.post import Post
from kilroy_face_debug.posters import Poster, BasicPoster
from kilroy_face_debug.processors import Processor
from kilroy_face_debug.restrictions import Restriction
from kilroy_face_debug.scoring.modifiers import ScoreModifier
from kilroy_face_debug.scoring.raw import Scorer, RandomScorer
from kilroy_face_debug.scrapers import Scraper, DummyScraper
from kilroy_face_server_py_sdk import (
    Categorizable,
    CategorizableBasedParameter,
    Face,
    JSONSchema,
    Metadata,
    Parameter,
    Savable,
    SerializableModel,
    classproperty,
    normalize,
)
from kilroy_server_py_utils import (
    CategorizableBasedOptionalParameter,
    Configurable,
)

logger = logging.getLogger(__name__)


class Params(SerializableModel):
    poster_type: str = "basic"
    posters_params: Dict[str, Dict[str, Any]] = {}
    scorer_type: str = "random"
    scorers_params: Dict[str, Dict[str, Any]] = {}
    score_modifier_type: Optional[str] = None
    score_modifiers_params: Dict[str, Dict[str, Any]] = {}
    scraper_type: str = "dummy"
    scrapers_params: Dict[str, Dict[str, Any]] = {}
    restriction_type: Optional[str] = None
    restrictions_params: Dict[str, Dict[str, Any]] = {}


@dataclass
class State:
    processor: Processor
    poster: Poster
    posters_params: Dict[str, Dict[str, Any]]
    scorer: Scorer
    scorers_params: Dict[str, Dict[str, Any]]
    score_modifier: Optional[ScoreModifier]
    score_modifiers_params: Dict[str, Dict[str, Any]]
    scraper: Scraper
    scrapers_params: Dict[str, Dict[str, Any]]
    restriction: Optional[Restriction]
    restrictions_params: Dict[str, Dict[str, Any]]
    cache: Dict[UUID, Post]


class PosterParameter(CategorizableBasedParameter[State, Poster]):
    @classmethod
    async def _get_params(cls, state: State, category: str) -> Dict[str, Any]:
        return {**state.posters_params.get(category, {})}

    # noinspection PyMethodParameters
    @classproperty
    def default_categorizable(cls) -> Type[Poster]:
        return BasicPoster


class ScorerParameter(CategorizableBasedParameter[State, Scorer]):
    @classmethod
    async def _get_params(cls, state: State, category: str) -> Dict[str, Any]:
        return {**state.scorers_params.get(category, {})}

    # noinspection PyMethodParameters
    @classproperty
    def default_categorizable(cls) -> Type[Scorer]:
        return RandomScorer


class ScoreModifierParameter(
    CategorizableBasedOptionalParameter[State, ScoreModifier]
):
    @classmethod
    async def _get_params(cls, state: State, category: str) -> Dict[str, Any]:
        return {**state.score_modifiers_params.get(category, {})}


class ScraperParameter(CategorizableBasedParameter[State, Scraper]):
    @classmethod
    async def _get_params(cls, state: State, category: str) -> Dict[str, Any]:
        return {**state.scrapers_params.get(category, {})}

    # noinspection PyMethodParameters
    @classproperty
    def default_categorizable(cls) -> Type[Scraper]:
        return DummyScraper


class RestrictionParameter(
    CategorizableBasedOptionalParameter[State, Restriction]
):
    @classmethod
    async def _get_params(cls, state: State, category: str) -> Dict[str, Any]:
        return {**state.restrictions_params.get(category, {})}


class DebugFaceBase(Face[State], ABC):
    # noinspection PyMethodParameters
    @classproperty
    @abstractmethod
    def post_type(cls) -> str:
        pass

    @classmethod
    async def _build_processor(cls) -> Processor:
        return await cls._build_generic(Processor, category=cls.post_type)

    @classmethod
    async def _build_poster(cls, params: Params) -> Poster:
        return await cls._build_generic(
            Poster,
            category=params.poster_type,
            **params.posters_params.get(params.poster_type, {}),
        )

    @classmethod
    async def _build_scorer(cls, params: Params) -> Scorer:
        return await cls._build_generic(
            Scorer,
            category=params.scorer_type,
            **params.scorers_params.get(params.scorer_type, {}),
        )

    @classmethod
    async def _build_score_modifier(
        cls, params: Params
    ) -> Optional[ScoreModifier]:
        if params.score_modifier_type is None:
            return None
        return await cls._build_generic(
            ScoreModifier,
            category=params.score_modifier_type,
            **params.score_modifiers_params.get(
                params.score_modifier_type, {}
            ),
        )

    @classmethod
    async def _build_scraper(cls, params: Params) -> Scraper:
        return await cls._build_generic(
            Scraper,
            category=params.scraper_type,
            **params.scrapers_params.get(params.scraper_type, {}),
        )

    @classmethod
    async def _build_restriction(cls, params: Params) -> Optional[Restriction]:
        if params.restriction_type is None:
            return None
        return await cls._build_generic(
            Restriction,
            category=params.restriction_type,
            **params.restrictions_params.get(params.restriction_type, {}),
        )

    async def _build_default_state(self) -> State:
        params = Params(**self._kwargs)

        return State(
            processor=await self._build_processor(),
            poster=await self._build_poster(params),
            posters_params=params.posters_params,
            scorer=await self._build_scorer(params),
            scorers_params=params.scorers_params,
            score_modifier=await self._build_score_modifier(params),
            score_modifiers_params=params.score_modifiers_params,
            scraper=await self._build_scraper(params),
            scrapers_params=params.scrapers_params,
            restriction=await self._build_restriction(params),
            restrictions_params=params.restrictions_params,
            cache={},
        )

    @staticmethod
    async def _save_processor(state: State, directory: Path) -> None:
        if isinstance(state.processor, Savable):
            await state.processor.save(directory / "processor")

    @staticmethod
    async def _save_poster(state: State, directory: Path) -> None:
        if isinstance(state.poster, Savable):
            await state.poster.save(directory / "poster")

    @staticmethod
    async def _save_scorer(state: State, directory: Path) -> None:
        if isinstance(state.scorer, Savable):
            await state.scorer.save(directory / "scorer")

    @staticmethod
    async def _save_score_modifier(state: State, directory: Path) -> None:
        if isinstance(state.score_modifier, Savable):
            await state.score_modifier.save(directory / "score_modifier")

    @staticmethod
    async def _save_scraper(state: State, directory: Path) -> None:
        if isinstance(state.scraper, Savable):
            await state.scraper.save(directory / "scraper")

    @staticmethod
    async def _save_restriction(state: State, directory: Path) -> None:
        if isinstance(state.restriction, Savable):
            await state.restriction.save(directory / "restriction")

    @staticmethod
    async def _save_cache(state: State, directory: Path) -> None:
        serialized = {
            str(post_id): post.json() for post_id, post in state.cache.items()
        }
        with open(directory / "cache.json", "w") as f:
            json.dump(serialized, f)

    @staticmethod
    async def _create_state_dict(state: State) -> Dict[str, Any]:
        return {
            "processor_type": state.processor.category,
            "poster_type": state.poster.category,
            "scorer_type": state.scorer.category,
            "score_modifier_type": state.score_modifier.category
            if state.score_modifier is not None
            else None,
            "scraper_type": state.scraper.category,
            "restriction_type": state.restriction.category
            if state.restriction is not None
            else None,
            "posters_params": state.posters_params,
            "scorers_params": state.scorers_params,
            "score_modifiers_params": state.score_modifiers_params,
            "scrapers_params": state.scrapers_params,
            "restrictions_params": state.restrictions_params,
        }

    @classmethod
    async def _save_state(cls, state: State, directory: Path) -> None:
        await cls._save_processor(state, directory)
        await cls._save_poster(state, directory)
        await cls._save_scorer(state, directory)
        await cls._save_score_modifier(state, directory)
        await cls._save_scraper(state, directory)
        await cls._save_restriction(state, directory)
        await cls._save_cache(state, directory)
        state_dict = await cls._create_state_dict(state)
        await cls._save_state_dict(state_dict, directory)

    @classmethod
    async def _load_processor(
        cls, directory: Path, state_dict: Dict[str, Any]
    ) -> Processor:
        category = state_dict.get("processor_type", cls.post_type)
        return await cls._load_generic(
            directory / "processor",
            Processor,
            category=category,
            default=partial(cls._build_processor),
        )

    @classmethod
    async def _load_poster(
        cls, directory: Path, state_dict: Dict[str, Any], params: Params
    ) -> Poster:
        category = state_dict.get("poster_type", params.poster_type)
        return await cls._load_generic(
            directory / "poster",
            Poster,
            category=category,
            **state_dict.get("posters_params", params.posters_params).get(
                category, {}
            ),
            default=partial(cls._build_poster, params),
        )

    @classmethod
    async def _load_scorer(
        cls, directory: Path, state_dict: Dict[str, Any], params: Params
    ) -> Scorer:
        category = state_dict.get("scorer_type", params.scorer_type)
        return await cls._load_generic(
            directory / "scorer",
            Scorer,
            category=category,
            **state_dict.get("scorers_params", params.scorers_params).get(
                category, {}
            ),
            default=partial(cls._build_scorer, params),
        )

    @classmethod
    async def _load_score_modifier(
        cls, directory: Path, state_dict: Dict[str, Any], params: Params
    ) -> Optional[ScoreModifier]:
        if "score_modifier_type" in state_dict:
            category = state_dict["score_modifier_type"]
        else:
            category = params.score_modifier_type

        return await cls._load_generic(
            directory / "score_modifier",
            ScoreModifier,
            category=category,
            **state_dict.get(
                "score_modifiers_params", params.score_modifiers_params
            ).get(category, {}),
            default=partial(cls._build_score_modifier, params),
        )

    @classmethod
    async def _load_scraper(
        cls, directory: Path, state_dict: Dict[str, Any], params: Params
    ) -> Scraper:
        category = state_dict.get("scraper_type", params.scraper_type)
        return await cls._load_generic(
            directory / "scraper",
            Scraper,
            category=category,
            **state_dict.get("scrapers_params", params.scrapers_params).get(
                category, {}
            ),
            default=partial(cls._build_scraper, params),
        )

    @classmethod
    async def _load_restriction(
        cls, directory: Path, state_dict: Dict[str, Any], params: Params
    ) -> Optional[Restriction]:
        if "restriction_type" in state_dict:
            category = state_dict["restriction_type"]
        else:
            category = params.restriction_type

        if category is None:
            return None
        return await cls._load_generic(
            directory / "restriction",
            Restriction,
            category=category,
            **state_dict.get(
                "restrictions_params", params.restrictions_params
            ).get(category, {}),
            default=partial(cls._build_restriction, params),
        )

    @staticmethod
    async def _load_cache(directory: Path) -> Dict[UUID, Post]:
        with open(directory / "cache.json", "r") as f:
            serialized = json.load(f)
        return {
            UUID(post_id): Post(**json.loads(post))
            for post_id, post in serialized.items()
        }

    async def _load_saved_state(self, directory: Path) -> State:
        state_dict = await self._load_state_dict(directory)
        params = Params(**self._kwargs)

        return State(
            processor=await self._load_processor(directory, state_dict),
            poster=await self._load_poster(directory, state_dict, params),
            posters_params=state_dict.get(
                "posters_params", params.posters_params
            ),
            scorer=await self._load_scorer(directory, state_dict, params),
            scorers_params=state_dict.get(
                "scorers_params", params.scorers_params
            ),
            score_modifier=await self._load_score_modifier(
                directory, state_dict, params
            ),
            score_modifiers_params=state_dict.get(
                "score_modifiers_params", params.score_modifiers_params
            ),
            scraper=await self._load_scraper(directory, state_dict, params),
            scrapers_params=state_dict.get(
                "scrapers_params", params.scrapers_params
            ),
            restriction=await self._load_restriction(
                directory, state_dict, params
            ),
            restrictions_params=state_dict.get(
                "restrictions_params", params.restrictions_params
            ),
            cache=await self._load_cache(directory),
        )

    async def cleanup(self) -> None:
        async with self.state.write_lock() as state:
            if isinstance(state.processor, Configurable):
                await state.processor.cleanup()
            if isinstance(state.poster, Configurable):
                await state.poster.cleanup()
            if isinstance(state.scorer, Configurable):
                await state.scorer.cleanup()
            if isinstance(state.score_modifier, Configurable):
                await state.score_modifier.cleanup()
            if isinstance(state.scraper, Configurable):
                await state.scraper.cleanup()
            if isinstance(state.restriction, Configurable):
                await state.restriction.cleanup()


class DebugFace(DebugFaceBase, Categorizable, ABC):
    # noinspection PyMethodParameters
    @classproperty
    def category(cls) -> str:
        name: str = cls.__name__
        return normalize(name.removesuffix("DebugFace"))

    # noinspection PyMethodParameters
    @classproperty
    def metadata(cls) -> Metadata:
        return Metadata(
            key="kilroy-face-debug", description="Kilroy face for debugging"
        )

    # noinspection PyMethodParameters
    @classproperty
    def post_type(cls) -> str:
        return cls.category

    # noinspection PyMethodParameters
    @classproperty
    def post_schema(cls) -> JSONSchema:
        return Processor.for_category(cls.post_type).post_schema

    # noinspection PyMethodParameters
    @classproperty
    def parameters(cls) -> Set[Type[Parameter]]:
        return {
            PosterParameter,
            ScorerParameter,
            ScoreModifierParameter,
            ScraperParameter,
            RestrictionParameter,
        }

    async def post(
        self, content: Dict[str, Any]
    ) -> Tuple[UUID, Optional[str]]:
        logger.info("Creating new post...")

        async with self.state.read_lock() as state:
            data = await state.processor.to_internal(content)
            if state.restriction is not None:
                if not await state.restriction.check(data):
                    raise ValueError("Post is not allowed to be posted.")
            post = await state.poster.post(state.cache, data)

        logger.info(f"New post id: {str(post.id)}.")
        return post.id, None

    async def score(self, id: UUID) -> float:
        logger.info(f"Scoring post {str(id)}...")

        async with self.state.read_lock() as state:
            post = state.cache[id]
            score = await state.scorer.score(post)
            if state.score_modifier is not None:
                score = await state.score_modifier.modify(post, score)

        logger.info(f"Score for post {str(id)}: {score}.")
        return score

    async def _fetch(
        self,
        posts: AsyncIterable[Post],
    ) -> AsyncIterable[Tuple[UUID, Dict[str, Any], float]]:
        posts = aiter(posts)

        while True:
            async with self.state.read_lock() as state:
                try:
                    post = await anext(posts)
                except StopAsyncIteration:
                    break

                score = await state.scorer.score(post)
                if state.score_modifier is not None:
                    score = await state.score_modifier.modify(post, score)

                try:
                    data = await state.processor.to_external(post.data)
                except Exception:
                    continue

                yield post.id, data, score

    async def scrap(
        self,
        limit: Optional[int] = None,
        before: Optional[datetime] = None,
        after: Optional[datetime] = None,
    ) -> AsyncIterable[Tuple[UUID, Dict[str, Any], float]]:
        async with self.state.read_lock() as state:
            posts = state.scraper.scrap(before, after)

        posts = self._fetch(posts)
        if limit is not None:
            posts = stream.take(posts, limit)
        else:
            posts = stream.iterate(posts)

        logger.info("Scraping posts...")

        async with posts.stream() as streamer:
            async for post_id, post, score in streamer:
                logger.info(f"Scraped post {str(post_id)}.")
                yield post_id, post, score

        logger.info("Scraping finished.")

    async def reset_self(self) -> None:
        logger.info("Resetting state...")
        await super().reset_self()
        logger.info("State reset.")

    async def save_self(self, directory: Path) -> None:
        logger.info("Saving state...")
        await super().save_self(directory)
        logger.info("State saved.")


class TextOnlyDebugFace(DebugFace):
    pass


class ImageOnlyDebugFace(DebugFace):
    pass


class TextAndImageDebugFace(DebugFace):
    pass


class TextOrImageDebugFace(DebugFace):
    pass


class TextWithOptionalImageDebugFace(DebugFace):
    pass


class ImageWithOptionalTextDebugFace(DebugFace):
    pass



from abc import ABC, abstractmethod
from typing import MutableMapping
from uuid import UUID, uuid4

from kilroy_server_py_utils import Categorizable, classproperty, normalize

from kilroy_face_debug.post import PostData, Post


class Poster(Categorizable, ABC):
    # noinspection PyMethodParameters
    @classproperty
    def category(cls) -> str:
        name: str = cls.__name__
        return normalize(name.removesuffix("Poster"))

    @abstractmethod
    async def post(
        self, cache: MutableMapping[UUID, Post], data: PostData
    ) -> Post:
        pass


# Basic


class BasicPoster(Poster):
    async def post(
        self, cache: MutableMapping[UUID, Post], data: PostData
    ) -> Post:
        uid = uuid4()
        post = Post(data=data, id=uid)
        cache[uid] = post
        return post



import json
from abc import ABC, abstractmethod
from dataclasses import dataclass
from pathlib import Path
from typing import Dict, Any

from detoxify import Detoxify
from kilroy_face_py_shared import SerializableModel
from kilroy_face_server_py_sdk import Categorizable, classproperty, normalize
from kilroy_server_py_utils import Configurable, Parameter, background

from kilroy_face_debug.models import ToxicityModelLoader
from kilroy_face_debug.post import PostData


class Restriction(Categorizable, ABC):
    # noinspection PyMethodParameters
    @classproperty
    def category(cls) -> str:
        name: str = cls.__name__
        return normalize(name.removesuffix("Restriction"))

    @abstractmethod
    async def check(self, data: PostData) -> bool:
        pass


# Toxicity


class ToxicityRestrictionParams(SerializableModel):
    threshold: float = 0.8


@dataclass
class ToxicityRestrictionState:
    detoxify: Detoxify
    threshold: float


class ToxicityRestriction(Restriction, Configurable[ToxicityRestrictionState]):
    class ThresholdParameter(Parameter[ToxicityRestrictionState, float]):
        # noinspection PyMethodParameters
        @classproperty
        def schema(cls) -> Dict[str, Any]:
            return {
                "type": "number",
                "minimum": 0,
                "maximum": 1,
                "title": cls.pretty_name,
                "default": 0.8,
            }

    async def _build_default_state(self) -> ToxicityRestrictionState:
        params = ToxicityRestrictionParams(**self._kwargs)
        return ToxicityRestrictionState(
            detoxify=await background(ToxicityModelLoader.get),
            threshold=params.threshold,
        )

    @classmethod
    async def _save_state(
        cls, state: ToxicityRestrictionState, directory: Path
    ) -> None:
        state_dict = {"threshold": state.threshold}
        with open(directory / "state.json", "w") as f:
            json.dump(state_dict, f)

    async def _load_saved_state(
        self, directory: Path
    ) -> ToxicityRestrictionState:
        params = ToxicityRestrictionState(**self._kwargs)
        with open(directory / "state.json", "r") as f:
            state_dict = json.load(f)
        return ToxicityRestrictionState(
            detoxify=await background(ToxicityModelLoader.get),
            threshold=state_dict.get("threshold", params.threshold),
        )

    async def cleanup(self) -> None:
        await background(ToxicityModelLoader.release)

    async def check(self, data: PostData) -> bool:
        if data.text is None:
            return True
        async with self.state.read_lock() as state:
            toxicity = state.detoxify.predict(data.text.content)["toxicity"]
            return toxicity < state.threshold



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['kilroy_face_debug',
 'kilroy_face_debug.resources',
 'kilroy_face_debug.scoring']

package_data = \
{'': ['*']}

install_requires = \
['detoxify>=0.5,<0.6',
 'kilroy-face-server-py-sdk>=0.9,<0.10',
 'numpy>=1.23,<2.0',
 'omegaconf>=2.2,<3.0',
 'platformdirs>=2.5,<3.0',
 'pydantic[dotenv]>=1.10,<2.0',
 'typer[all]>=0.6,<0.7']

entry_points = \
{'console_scripts': ['kilroy-face-debug = kilroy_face_debug.__main__:cli',
                     'kilroy-face-debug-fetch-models = '
                     'kilroy_face_debug.models:fetch_models']}

setup_kwargs = {
    'name': 'kilroy-face-debug',
    'version': '0.1.0',
    'description': 'kilroy face for Debug 🎮',
    'long_description': '<h1 align="center">kilroy-face-debug</h1>\n\n<div align="center">\n\nkilroy face for debugging 🧪\n\n[![Lint](https://github.com/kilroybot/kilroy-face-debug/actions/workflows/lint.yaml/badge.svg)](https://github.com/kilroybot/kilroy-face-debug/actions/workflows/lint.yaml)\n[![Multiplatform tests](https://github.com/kilroybot/kilroy-face-debug/actions/workflows/test-multiplatform.yaml/badge.svg)](https://github.com/kilroybot/kilroy-face-debug/actions/workflows/test-multiplatform.yaml)\n[![Docker tests](https://github.com/kilroybot/kilroy-face-debug/actions/workflows/test-docker.yaml/badge.svg)](https://github.com/kilroybot/kilroy-face-debug/actions/workflows/test-docker.yaml)\n[![Docs](https://github.com/kilroybot/kilroy-face-debug/actions/workflows/docs.yaml/badge.svg)](https://github.com/kilroybot/kilroy-face-debug/actions/workflows/docs.yaml)\n\n</div>\n\n---\n\n## Installing\n\nUsing `pip`:\n\n```sh\npip install kilroy-face-debug\n```\n\n## Usage\n\nTo run the face server, install the package and run the following command:\n\n```sh\nkilroy-face-debug\n```\n\nThis will start the face server on port 9999 by default.\nThen you can communicate with the server, for example by using\n[this package](https://github.com/kilroybot/kilroy-face-client-py-sdk).\n',
    'author': 'kilroy',
    'author_email': 'kilroymail@pm.me',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/kilroybot/kilroy-face-debug',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.10,<3.11',
}


setup(**setup_kwargs)



from typing import Optional
from uuid import UUID

from kilroy_face_py_shared import SerializableModel


class PostTextData(SerializableModel):
    content: str


class PostImageData(SerializableModel):
    raw: str
    filename: Optional[str]


class PostData(SerializableModel):
    text: Optional[PostTextData]
    image: Optional[PostImageData]


class Post(SerializableModel):
    data: PostData
    id: UUID
